﻿using RaDynamics.QrZahlteilApp.Service;
using System;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace QrZahlteilExample
{
    public partial class MainForm : Form
    {
        private QrZahlteilServiceApi _api;
        private FailoverProvider _failoverProvider;
        private int _connectionFailedCounter;
        private string _sessionId;
        private byte[] _lastPdf;

        public MainForm()
        {
            InitializeComponent();
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            lblTitel.Text += $@" Version {Assembly.GetExecutingAssembly().GetName().Version}";

            _api_ScannerDisconnected(null, null);
            EnableButtons();
        }

        private void cmdVerbinden_Click(object sender, EventArgs e)
        {
            cmdVerbinden.Enabled = false;

            CreateApi();

            _api.PictureFormat = PictureFormat.NoPictures;
            if (rdoPictureFormatPdf.Checked) {
                _api.PictureFormat = PictureFormat.Pdf;
            } else if (rdoPictureFormatSeparatePictures.Checked) {
                _api.PictureFormat = PictureFormat.SeparatePictures;
            }

            if (rdoEncryptionE2ee.Checked) {
                // Bei eingesetzer Ende-zu-Ende Verschlüsselung muss der QR-Code zur Kopplung zwischen Ihrer Applikation und der QR-Zahlteil App selbst generiert werden.
                // Die Werte secret/iv müssen pro Anwender generiert werden.
                var secret = Encoding.UTF8.GetBytes(txtSecret.Text);
                var iv = Encoding.UTF8.GetBytes(txtIV.Text);
                _api.SetEncryption(new AesEncryption(secret, iv), new QrCodeGenerator());
            } else {
                // Keine Verschlüsselung der Nutzdaten. Die Daten werden TLS verschlüsselt übertragen.
                _api.SetEncryption(null, null);
            }

            EnableOptions(false);
            Task.Run(() =>
            {
                if (!_api.IsClosed) {
                    return;
                }
                _api.Start(txtGeraeteName.Text);
            });
        }

        private void CreateApi()
        {
            if (_api != null) {
                if (!_api.IsClosed) {
                    _api.Stop(false);
                }
                _api = null;
            }

            // Speichern und Übergabe der sessionId beim Verbinden ermöglichst eine automatische Kopplung (Pairing) zwischen Scanner und diesem Programm.
            _api = new QrZahlteilServiceApi(txtApiKey.Text, _sessionId)
            {
                IdentificationKey = txtIdentificationKey.Text
            };
            _failoverProvider = new FailoverProvider(_api.ServiceUri);
            // Wenn nach einigen Versuchen keine Verbindung möglich ist, könnte der produktive Server nicht verfügbar sein. In diesem Fall wird versucht,
            // auf einen alternativen Server zu wechseln. Bei einer Änderung der Uri muss der Anwender die Kopplung zwischen Smartphone und Ihrem Programm
            // erneut durchführen. Nur dadurch verbindet auch die Smartphone App auf die neue Uri.
            if (_connectionFailedCounter >= 3) {
                // Speichern Sie diese Serverinformationen nicht. Verbinden Sie nach jedem Programmstart zuerst erneut auf den regulären produtiven Server.
                var server = _failoverProvider.Get(_api.ServiceUri);
                _api.ServiceUri = server.Uri;
                // Einige Ausfallserver benötigen ein KeepAlive Signal.
                _api.KeepAlive = server.KeepAlive;
            }
            _api.Connected += _api_Connected;
            _api.Close += _api_Close;
            _api.ScannerConnected += _api_ScannerConnected;
            _api.ScannerDisconnected += _api_ScannerDisconnected;
            _api.Scanned += _api_Scanned;
            _api.StartFailed += _api_StartFailed;
            _api.ServerError += _api_ServerError;
        }

        private void _api_Connected(object sender, ConnectedEventArgs e)
        {
            BeginInvoke(new MethodInvoker(() =>
            {
                _connectionFailedCounter = 0;
                lblStatus.Text = "Verbindung zu Service hergestellt";
                // Der Service kann neue sessionId liefern. Die hier gelieferte soll für eine automatische Kopplung (Pairing) gespeichert werden.
                _sessionId = e.SessionId;
                picPairingQrCode.Image = e.QrCode;
                picPairingQrCode.Visible = true;
                EnableButtons();
            }));
        }

        private void _api_Close(object sender, EventArgs e)
        {
            BeginInvoke(new MethodInvoker(() =>
            {
                lblStatus.Text = "Verbindung zu Service geschlossen";
                EnableOptions(true);
                EnableButtons();
                picPairingQrCode.Visible = false;
            }));
        }

        private void _api_ScannerConnected(object sender, ScannerConnectedEventArgs e)
        {
            BeginInvoke(new MethodInvoker(() =>
            {
                lblStatus.Text = $"Verbunden mit {e.ScannerName}";
                picPairingQrCode.Visible = false;
            }));
        }

        private void _api_ScannerDisconnected(object sender, EventArgs e)
        {
            BeginInvoke(new MethodInvoker(() =>
            {
                lblStatus.Text = "Smartphone nicht verbunden";
                picPairingQrCode.Visible = _api != null;
                EnableButtons();
            }));
        }

        private void _api_Scanned(object sender, ScannedEventArgs e)
        {
            BeginInvoke(new MethodInvoker(() =>
            {
                txtQrCode.Text = NormalizeLinebreaks(e.Data.Raw);
                pnlPictures.Controls.Clear();
                for (int i = e.Data.Pictures.Length - 1; i >= 0; i--) {
                    pnlPictures.Controls.Add(new PictureBox
                    {
                        Dock = DockStyle.Top,
                        Size = new Size(pnlPictures.Width, 400),
                        SizeMode = PictureBoxSizeMode.Zoom,
                        Image = e.Data.Pictures[i]
                    });
                }

                cmdOpenPdf.Visible = e.Data.HasPdf;
                _lastPdf = e.Data.Pdf;
            }));
        }

        private void _api_StartFailed(object sender, StartFailedEventArgs e)
        {
            BeginInvoke(new MethodInvoker(() =>
            {
                lblStatus.Text = $"Verbindung zu {_api.ServiceUri} konnte nicht hergestellt werden.";
                _connectionFailedCounter++;
                EnableButtons();
            }));
        }

        private void _api_ServerError(object sender, ServerErrorEventArgs e)
        {
            BeginInvoke(new MethodInvoker(() =>
            {
                lblStatus.Text = $"ServerError {e.ErrorCode} {e.Key} {e.Text ?? string.Empty}";
                EnableButtons();
            }));
        }

        private void cmdTrennen_Click(object sender, EventArgs e)
        {
            _api.Stop(false);
        }

        private void cmdResetSessionId_Click(object sender, EventArgs e)
        {
            _sessionId = null;
            if (_api == null || _api.IsClosed) {
                return;
            }
            _api.Stop(true);
        }

        private void EnableOptions(bool enabled)
        {
            txtGeraeteName.Enabled = enabled;
            txtApiKey.Enabled = enabled;
            rdoEncryptionNone.Enabled = enabled;
            rdoEncryptionE2ee.Enabled = enabled;
            txtSecret.Enabled = enabled && rdoEncryptionE2ee.Checked;
            txtIV.Enabled = enabled && rdoEncryptionE2ee.Checked;
            txtIdentificationKey.Enabled = enabled;
            rdoPictureFormatNone.Enabled = enabled;
            rdoPictureFormatPdf.Enabled = enabled;
            rdoPictureFormatSeparatePictures.Enabled = enabled;
        }

        private void EnableButtons()
        {
            cmdVerbinden.Enabled = _api == null || _api.IsClosed;
            cmdTrennen.Enabled = !cmdVerbinden.Enabled;
        }

        private string NormalizeLinebreaks(string text)
        {
            return Regex.Replace(text, @"\r\n|\n\r|\n|\r", "\r\n");
        }

        private void lblWebsite_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            Process.Start(lblWebsite.Text);
        }

        private void lblApiKey_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            Process.Start("https://www.qrzahlteil.ch/api.html");
        }

        private void cmdOpenPdf_Click(object sender, EventArgs e)
        {
            var fileName = Path.Combine(Path.GetTempPath(), $"{Path.GetRandomFileName()}.pdf");
            using (FileStream stream = File.Create(fileName)) {
                stream.Write(_lastPdf, 0, _lastPdf.Length);
            }
            Process.Start(fileName);
        }

        private void rdoEncryption_CheckedChanged(object sender, EventArgs e)
        {
            txtSecret.Enabled = rdoEncryptionE2ee.Checked;
            txtIV.Enabled = rdoEncryptionE2ee.Checked;
        }
    }
}
