﻿using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;

namespace RaDynamics.QrZahlteilApp.Service
{
    /// <summary>
    /// Diese Klasse kann alternative Service Uri liefern, sofern der produktive Server temporär nicht erreichbar ist.
    /// <remarks>Alternative Service Uri sind variabel und nicht ständig erreichbar. Versuchen Sie immer zuerst den produktiven Server zu erreichen.</remarks>
    /// </summary>
    public class FailoverProvider
    {
        private readonly Server _defaultServer;

        /// <summary>
        /// Fallback, falls die Onlinequelle für alternative Service Uri nicht verfügbar ist.
        /// <remarks>Es sollte immer zuerst die aktuelle Onlinequelle geprüft werden, da diese statische Auflistung veraltet sein kann.</remarks>
        /// </summary>
        private static readonly Server[] FallbackServers = {
            new Server(new Uri("wss://qrzahlteilapp.azurewebsites.net/"), TimeSpan.FromSeconds(60)),
            new Server(new Uri("wss://failover.qrzahlteil.ch/api/"), TimeSpan.Zero)
        };

        /// <summary>
        /// Zu verwendender HTTP Proxyserver oder null, falls die Systemeinstellungen verwendet werden sollen.
        /// </summary>
        public IWebProxy Proxy { get; set; }

        /// <summary>
        /// Konstruktor mit normalerweise verwendeter Uri (<param name="defaultServiceUri"></param>
        /// </summary>
        /// <param name="defaultServiceUri">Normalerweise verwendete Uri (meist <see cref="QrZahlteilServiceApi.DefaultServiceUri"/>)</param>
        public FailoverProvider(Uri defaultServiceUri)
        {
            if (defaultServiceUri == null) throw new ArgumentNullException(nameof(defaultServiceUri));
            _defaultServer = new Server(defaultServiceUri, TimeSpan.Zero);
        }

        public Server Get(Uri current)
        {
            var servers = GetServerFromWebSource() ?? new List<Server>(FallbackServers);
            if (servers.Count == 0) {
                return _defaultServer;
            }

            servers.Insert(0, _defaultServer);

            // Currently used entry may not included (ex. testing)
            var indexCurrent = Array.IndexOf(servers.Select(o => o.Uri).ToArray(), current);
            // Currently used last item -> take first entry again
            var usedLast = indexCurrent + 1 >= servers.Count;
            if (indexCurrent == -1 || usedLast) {
                return servers.First();
            }

            return servers[indexCurrent + 1];
        }

        private List<Server> GetServerFromWebSource()
        {
            var request = (HttpWebRequest)WebRequest.Create("https://www.radynamics.com/qrzahlteilfailover/");
            request.Timeout = 10000;
            request.Proxy = Proxy ?? request.Proxy;
            try {
                var response = (HttpWebResponse)request.GetResponse();
                using (var sr = new StreamReader(response.GetResponseStream())) {
                    var result = sr.ReadToEnd();
                    var json = JArray.Parse(result);

                    if (json.Count == 0) {
                        return new List<Server>();
                    }

                    var servers = new List<Server>();
                    for (int i = 0; i < json.Count; i++) {
                        servers.Add(CreateServer((JObject)json[i]));
                    }

                    return servers;
                }
            } catch {
                // WebException occures if no internet connection is available
                // WebException thrown on timeout
                // Ignore all other issues (ex. malformed json)
            }

            return null;
        }

        private static Server CreateServer(JObject json)
        {
            var uri = new Uri(json["uri"].Value<string>());
            var keepAlive = TimeSpan.Zero; ;
            if (json.ContainsKey("keepAliveIntervalSecs")) {
                long value = json["keepAliveIntervalSecs"].Value<long>();
                keepAlive = value > 0 ? TimeSpan.FromSeconds(value) : TimeSpan.Zero;
            }
            return new Server(uri, keepAlive);
        }
    }
}
