﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Threading;
using System.Threading.Tasks;
using WebSocketSharp;

namespace RaDynamics.QrZahlteilApp.Service
{
    public class QrZahlteilServiceApi
    {
        private WebSocket _client;
        private Uri _serviceUri = DefaultServiceUri;
        private readonly string _apiKey;
        private string _sessionId;
        private System.Timers.Timer _keepAliveTimer;

        private IQrCodeGenerator _qrCodeGenerator;

        /// <summary>
        /// Definiert wie vom Benutzer aufgenommene Bilder der Rechnung von der Applikation empfangen werden. Bei NoPictures macht der Benutzer keine Fotos der Rechnung.
        /// <remarks>Muss vor <see cref="Start()"/> gesetzt werden oder die Verbindung muss geschlossen und erneut geöffnet werden.</remarks>
        /// </summary>
        public PictureFormat PictureFormat { get; set; }

        /// <summary>
        /// Das Verfahren zur Ver-/ und Entschlüsselung von Nutzdaten oder <seealso cref="NoEncryption"/> falls keine Verschlüsselung der Nutzdaten stattfindet.
        /// </summary>
        public IEncryption Encryption { get; private set; } = new NoEncryption();

        /// <summary>
        /// Lizenzschlüssel für Endanwender. Für Ihre Applikation kann dieser Parameter weggelassen werden, bzw. beim Standardwert belassen werden.
        /// </summary>
        public string LicenseKey { get; set; }

        /// <summary>
        /// Zeitpunkt bis wann Endanwender diesen Dienst als Vollversion nutzen können.
        /// </summary>
        public DateTime LicensedUntil { get; private set; }

        /// <summary>
        /// True, falls die Nutzdaten verschlüsselt übertragen werden.
        /// <remarks>Der QR-Code zur Kopplung wird dabei nicht vom Server als Bild geliefert. Das Bild muss mit <see cref="IQrCodeGenerator"/> manuell erstellt werden.</remarks>
        /// </summary>
        public bool IsEncrypted => _qrCodeGenerator != null;
        /// <summary>
        /// Key als Freitext zur Identifikation einer Installation oder eines Anwenders, welcher zu jedem Scan im Klartext gespeichert wird.
        /// </summary>
        public string IdentificationKey { get; set; }

        /// <summary>
        /// Adresse des zu verwendenden HTTP Proxyserver oder null, falls kein Proxyserver verwendet wird.
        /// </summary>
        public Uri ProxyServerUrl { get; set; }
        /// <summary>
        /// Benutzername für den HTTP Proxyserver oder null, falls kein Benutzername verwendet wird.
        /// </summary>
        public string ProxyServerUsername { get; set; }
        /// <summary>
        /// Passwort für den HTTP Proxyserver oder null, falls kein Passwort verwendet wird.
        /// </summary>
        public string ProxyServerPassword { get; set; }

        /// <summary>
        /// Standard Uri zum Backend der QR-Zahlteil App.
        /// </summary>
        public static readonly Uri DefaultServiceUri = new Uri("wss://qrzahlteil.ch/api1/");

        /// <summary>
        /// Uri zum Backend der QR-Zahlteil App.
        /// <remarks>Bei einer Änderung dieses Wertes muss der Anwender die Kopplung zwischen Smartphone und Ihrem Programm erneut durchführen. Nur dadurch verbindet auch die Smartphone App auf die neue Uri.</remarks>
        /// </summary>
        public Uri ServiceUri
        {
            get => _serviceUri;
            set => _serviceUri = value ?? throw new ArgumentNullException(nameof(ServiceUri));
        }

        /// <summary>
        /// Standard KeepAlive zum Backend der QR-Zahlteil App.
        /// </summary>
        public static readonly TimeSpan DefaultKeepAlive = TimeSpan.Zero;

        /// <summary>
        /// Intervall, in welchem ein noop-Signal an das Backend geschickt wird. Wenn negativ oder TimeSpan.Zero wird kein Signal geschickt.
        /// <remarks>Einige Backend-/Ausfallserver (z.B. azurewebsites.net) unterbrechen bei Inaktivität die Verbindung ohne OnClose-Event.</remarks>
        /// </summary>
        public TimeSpan KeepAlive { get; set; } = DefaultKeepAlive;

        /// <summary>
        /// Die verwendete Protokollversion dieser API.
        /// <remarks>Es sollte immer die aktuellste Version verwendet werden. Ältere Versionen unterliegen Einschränkungen und unterstützen teilweise nur einen reduzierten Funktionsumfang.</remarks>
        /// </summary>
        public int Version => 2;

        /// <summary>
        /// Erstellt einen Zugriff zum Backend der QR-Zahlteil App.
        /// </summary>
        /// <param name="apiKey">ApiKey der Applikation zur Identifikation. Ein neuer ApiKey kann über <see cref="https://www.qrzahlteil.ch/api.html"/>https://www.qrzahlteil.ch/api.html generiert werden.</param>
        /// <param name="sessionId">Gespeicherte sessionId oder leer ""</param>
        public QrZahlteilServiceApi(string apiKey, string sessionId)
            : this(apiKey, sessionId, DefaultServiceUri)
        {
        }

        /// <summary>
        /// Erstellt einen Zugriff zum Backend der QR-Zahlteil App
        /// </summary>
        /// <param name="apiKey">ApiKey der Applikation zur Identifikation. Ein neuer ApiKey kann über <see cref="https://www.qrzahlteil.ch/api.html"/>https://www.qrzahlteil.ch/api.html generiert werden.</param>
        /// <param name="sessionId">Gespeicherte sessionId oder leer ""</param>
        /// <param name="uri">Pfad zum Backend (aktuell "wss://qrzahlteil.ch/api1/")</param>
        public QrZahlteilServiceApi(string apiKey, string sessionId, Uri uri)
        {
            _apiKey = apiKey ?? throw new ArgumentNullException(nameof(apiKey));
            _sessionId = sessionId;
            ServiceUri = uri;
        }

        /// <summary>
        /// True, wenn die Verbindung geschlossen ist.
        /// </summary>
        public bool IsClosed => _client == null || !_client.IsAlive;

        /// <summary>
        /// True, wenn aktuell eine Verbindung hergestellt wird.
        /// </summary>
        public bool IsConnecting { get; private set; }

        private void _client_OnClose(object sender, CloseEventArgs e)
        {
            OnClose(EventArgs.Empty);
        }

        /// <summary>
        /// Behandelt vom Server empfangene Antworten und Meldungen
        /// </summary>
        private void _client_OnMessage(object sender, MessageEventArgs e)
        {
            var msg = JObject.Parse(e.Data);
            if (msg.ContainsKey("error")) {
                var error = (JObject)msg["error"];
                var text = GetErrorText(error);
                OnServerError(new ServerErrorEventArgs(error["code"].ToString(), error["key"].ToString(), text));
                return;
            }

            if (!msg.ContainsKey("response")) {
                return;
            }

            var response = (JObject)msg["response"];
            if (!response.ContainsKey("event")) {
                return;
            }

            var eventName = response["event"].ToString();

            // Die SmartphoneApp antwortet normalerweise in der Protokollversion, welche bei "connect" übergeben wurde. Wenn eine aktuelle Version der SmartphoneApp mit einer
            // zu alten Version Ihrer Applikation verwendet wird, ist dies evtl. nicht mehr möglich. Der Anwender sollte in diesem Fall auf ein notwendiges Update aufmerksam gemacht werden.
            var responseVersion = int.Parse(response["version"].ToString());
            if (responseVersion != Version) {
                OnServerError(new ServerErrorEventArgs("0", "0", $"Bei '{eventName}' wurde die Datenstruktur in der Version {Version} erwartet aber {responseVersion} geliefert. Bitte aktualisieren Sie Ihre empfangende Applikation."));
                return;
            }

            // Es wurde erfolgreich eine Verbindung zum Server hergestellt
            if (eventName == "connected") {
                // Speichern Sie die [sessionId] und übergeben Sie diese bei künftigen Verbindungen an "connect". Damit wird für den Benutzer eine automatische Verbindung
                // zwischen Ihrem Programm und dem Smartphone ermöglicht.
                _sessionId = response["sessionId"].ToString();
                // Zeitpunkt bis wann Endanwender diesen Dienst als Vollversion nutzen können.
                LicensedUntil = DateTime.Parse(response["licensedUntil"].ToString());

                // [custom] kann spezifische Parameter liefern. Falls für Ihren ApiKey keine solchen definiert wurden, ist diese Auflistung leer.
                var customParameter = new Dictionary<string, string>();
                var customChildren = response["custom"]?.Children() ?? (IEnumerable<JToken>)Array.Empty<JToken>();
                foreach (var c in customChildren) {
                    if (c is JProperty p) {
                        customParameter.Add(p.Name, p.Value.Value<string>());
                    }
                }

                Image pairingQrCodeImg;
                if (IsEncrypted) {
                    // [pairingQrCodeData] enthält die Daten, mit welchen ein eigenes [pairingQrCodeImg] generiert werden muss, wenn [secret] definiert ist
                    // und damit Ende-zu-Ende Verschlüsselung eingesetzt wird.
                    var qrCodeContent = JObject.Parse(response["pairingQrCodeData"].ToString());
                    // Damit nur das Smartphone den verwendeten Schlüssel/Secret kennt, muss dieses manuell dem QR-Code zur Kopplung hingefügt werden. Die Werte werden als Hex übergeben.
                    qrCodeContent["secret"] = Utils.ToHex(Encryption.Secret);
                    qrCodeContent["iv"] = Utils.ToHex(Encryption.IV);
                    pairingQrCodeImg = _qrCodeGenerator.Create(qrCodeContent.ToString());
                } else {
                    // [pairingQrCodeImg] enthält ein Base64 kodiertes Bild als PNG, welches dem Benutzer zur Kopplung zwischen Smartphone und Ihrem Programm angezeigt werden muss.
                    // Es wird nur geliefert, wenn [secret] undefiniert ist und damit keine Ende-zu-Ende Verschlüsselung eingesetzt wird.
                    pairingQrCodeImg = Utils.ToImage(response["pairingQrCodeImg"].ToString());
                }
                OnConnected(new ConnectedEventArgs(_sessionId, pairingQrCodeImg, customParameter));
                return;
            }

            // Der Benutzer hat das Smartphone erfolgreich mit Ihrem Programm verbunden
            if (eventName == "scanner_connected") {
                // Bei eingesetzter Ende-zu-Ende Verschlüsselung müssen erhaltene Daten vorgängig entschlüsselt werden.
                var data = GetPlaintextData(response["data"]);
                // [scannerName] enthält den Base64 kodierten Namen des Smartphones.
                OnScannerConnected(new ScannerConnectedEventArgs(Utils.FromBase64(data["scannerName"].ToString())));
                return;
            }

            // Der Benutzer hat eine QR-Rechnung erfolgreich gescannt
            if (eventName == "scanned") {
                // Bei eingesetzter Ende-zu-Ende Verschlüsselung müssen erhaltene Daten vorgängig entschlüsselt werden.
                var data = GetPlaintextData(response["data"]);

                // [type] enthält den erkannten Typ der gescannten Daten
                var type = data["type"].ToString();
                // [data_base64] enthält den Base64 kodierten QR-Zahlteil der QR-Rechnung
                var raw = Utils.FromBase64(data["data_base64"].ToString());

                var qrData = new ScannedData(ScannedTypeHelper.Convert(type), raw);
                // [pictures_base64] enthält eine Liste von Base64 kodierten Bildern als JPEG, welche der Benutzer von der QR-Rechnung erstellt hat. Wenn bei "connect" Ihre Applikation
                // mit dem Server [pictureFormat] nicht definiert wurde oder null war, wird eine leerer Auflistung geliefert.
                var pics = (JArray)data["pictures_base64"];
                foreach (var pic in pics) {
                    qrData.Add(Utils.ToImage(pic.ToString()));
                }

                // [pdf_base64] enthält ein als Base64 kodiertes PDF der aufgenommenen Bilder, wenn bei "connect" für [pictureFormat] "pdf" definiert wurde.
                if (data["pdf_base64"] != null) {
                    qrData.Pdf = Convert.FromBase64String(data["pdf_base64"].ToString());
                }
                OnScanned(new ScannedEventArgs(qrData));
                return;
            }

            // Der Benutzer hat die App auf dem Smartphone geschlossen
            if (eventName == "scanner_disconnected") {
                OnScannerDisconnected(EventArgs.Empty);
            }
        }

        private static string GetErrorText(JObject error)
        {
            var key = error["key"].ToString();
            switch (key) {
                case "scanner_encryption_mismatch":
                    return QrZahlteilServiceRes.ScannerEncryptionMismatch;
                case "licensekey_in_use":
                    return QrZahlteilServiceRes.LisenceKeyInUse;
                default:
                    return error.ContainsKey("text") ? error["text"].ToString() : null;
            }
        }

        /// <summary>
        /// Liefert [data] als JSON Objekt im Klartext.
        /// </summary>
        private JObject GetPlaintextData(JToken data)
        {
            // Bei eingesetzter Ende-zu-Ende Verschlüsselung müssen erhaltene Daten vorgängig entschlüsselt werden.
            return IsEncrypted
                ? JObject.Parse(Encryption.Decrypt(Utils.FromBase64String(data.ToString())))
                : JObject.Parse(data.ToString());
        }

        /// <summary>
        /// Stellt eine Verbindung zum Service her.
        /// </summary>
        public void Start()
        {
            Start(Environment.MachineName);
        }

        /// <summary>
        /// Stellt eine Verbindung zum Service her.
        /// <param name="deviceName">Gerätename oder Programm, welches dem Anwender auf dem Scanner als "verbunden mit" angezeigt wird</param>
        /// </summary>
        public void Start(string deviceName)
        {
            Start(deviceName, typeof(QrZahlteilServiceApi).Assembly.GetName().Version.ToString());
        }

        /// <summary>
        /// Stellt eine Verbindung zum Service her.
        /// <param name="deviceName">Gerätename oder Programm, welches dem Anwender auf dem Scanner als "verbunden mit" angezeigt wird</param>
        /// <param name="productVersion">Version der eingesetzten Software oder null, falls diese unbekannt ist.</param>
        /// </summary>
        public void Start(string deviceName, string productVersion)
        {
            if (_client != null && _client.ReadyState == WebSocketState.Open) {
                throw new QrZahlteilServiceApiException(QrZahlteilServiceRes.AlreadyConnectionOpen);
            }

            _client = new WebSocket(ServiceUri.AbsoluteUri);
            _client.Log.Output = (data, s) => { Debug.WriteLine(data); };

            // Ohne Angabe wird TLS1 (obsolet) verwendet (https://github.com/sta/websocket-sharp/issues/356).
            _client.SslConfiguration.EnabledSslProtocols = System.Security.Authentication.SslProtocols.None | System.Security.Authentication.SslProtocols.Tls12;

            SetProxyServer();

            _client.OnMessage -= _client_OnMessage;
            _client.OnClose -= _client_OnClose;

            IsConnecting = true;
            var ts = new CancellationTokenSource();
            var t = new Task(() =>
            {
                while (_client.ReadyState != WebSocketState.Open && !ts.Token.IsCancellationRequested) {
                    _client.Connect();
                    Thread.Sleep(1000);
                }
            }, ts.Token);
            t.Start();
            var timeout = new TimeSpan(0, 0, 10);

            void StartFailed(Exception ex)
            {
                IsConnecting = false;
                ts.Cancel();
                _client.Close();
                OnStartFailed(new StartFailedEventArgs(ex));
            }
            try {
                t.Wait(timeout);
            } catch (Exception ex) {
                StartFailed(ex);
                return;
            }

            if (t.Status != TaskStatus.RanToCompletion) {
                StartFailed(new QrZahlteilServiceApiException(string.Format(QrZahlteilServiceRes.ConnectTimeout, timeout)));
                return;
            }

            IsConnecting = false;
            _client.OnMessage += _client_OnMessage;
            _client.OnClose += _client_OnClose;

            Register(deviceName, productVersion);

            StopKeepAlive();
            if (KeepAlive.TotalSeconds > 0) {
                StartKeepAlive();
            }
        }

        private void StopKeepAlive()
        {
            if (_keepAliveTimer != null) {
                _keepAliveTimer.Stop();
                _keepAliveTimer = null;
            }
        }

        private void StartKeepAlive()
        {
            _keepAliveTimer = new System.Timers.Timer(KeepAlive.TotalMilliseconds);
            _keepAliveTimer.AutoReset = true;
            _keepAliveTimer.Elapsed += delegate
            {
                if (IsClosed) {
                    return;
                }

                var obj = new
                {
                    method = "ping",
                    // Verwendete Version der Parameterstruktur
                    version = Version,
                    data = "empty",
                };
                _client.Send(JsonConvert.SerializeObject(obj));
            };
            _keepAliveTimer.Start();
        }

        private void SetProxyServer()
        {
            if (ProxyServerUrl == null) {
                _client.SetProxy(null, null, null);
            } else {
                _client.SetProxy(ProxyServerUrl.OriginalString, ProxyServerUsername, ProxyServerPassword);
            }
        }

        private void Register(string deviceName, string productVersion)
        {
            var data = new
            {
                deviceName = Utils.ToBase64(deviceName)
            };

            var obj = new
            {
                method = "connect",
                // Verwendete Version der Parameterstruktur
                version = Version,
                // Gespeicherte sessionId oder leer ""
                sessionId = _sessionId,
                // Ihr ApiKey (kann auf https://www.qrzahlteil.ch generiert werden)
                apiKey = _apiKey,
                // Key als Freitext zur Identifikation einer Installation oder eines Anwenders, welcher zu jedem Scan im Klartext gespeichert wird.
                identificationKey = IdentificationKey ?? string.Empty,
                // Format [separatePictures, pdf], falls der Benutzer die gesamte QR-Rechnung fotografieren soll und diese Fotos von Ihrer Applikation empfangen werden können.
                // null, falls keine Bilder aufgenommen werden sollen.
                pictureFormat = PictureFormatHelper.ToKey(PictureFormat),
                // True, falls die Daten zwischen Ihrer Applikation und der Smartphone App mit [secret] Ende-zu-Ende verschlüsselt werden sollen.
                encrypted = IsEncrypted,
                // Lizenzschlüssel für Endanwender. Für Ihre Applikation kann dieser Parameter weggelassen werden.
                licenseKey = LicenseKey,
                // Version der eingesetzten Software oder null, falls diese unbekannt ist.
                productVersion,
                data = IsEncrypted ? (object)Utils.ToBase64(Encryption.Encrypt(JsonConvert.SerializeObject(data))) : data
            };
            _client.Send(JsonConvert.SerializeObject(obj));
        }

        /// <summary>
        /// Schliesst die Verbindung zum Service.
        /// </summary>
        /// <param name="resetSessionId">True, wenn die sessionId (für automatisches Pairing) zurückgesetzt werden soll.</param>
        public void Stop(bool resetSessionId)
        {
            StopKeepAlive();
            _client?.Close();
            if (resetSessionId) {
                _sessionId = null;
            }
        }

        /// <summary>
        /// Setzt den Algorithmus für die Ver-/ und Entschlüsselung der übertragenen Nutzdaten.
        /// </summary>
        /// <param name="encryption">Algorithmus zur Ver-/ und Entschlüsselung oder null, falls keine Verschlüsselung verwendet wird</param>
        /// <param name="qrCodeGenerator"> Bei eingesetzter Verschlüsselung muss ein QR-Code Generator für die Erstellung des QR-Codes zur Kopplung angegeben werden.</param>
        /// <remarks>Wenn die Nutzdaten <see cref="IsEncrypted"/> sind, wird der QR-Code zur Kopplung nicht vom Server geliefert und muss manuell erstellt werden.</remarks>
        public void SetEncryption(IEncryption encryption, IQrCodeGenerator qrCodeGenerator)
        {
            if (encryption == null) {
                _qrCodeGenerator = null;
                Encryption = new NoEncryption();
            } else {
                _qrCodeGenerator = qrCodeGenerator ?? throw new ArgumentNullException(nameof(qrCodeGenerator));
                Encryption = encryption;
            }
        }

        #region Events
        /// <summary>
        /// Tritt auf, wenn ein Verbindungsaufbau zum Service nicht erfolgreich war.
        /// </summary>
        internal event EventHandler<StartFailedEventArgs> StartFailed
        {
            add => _startFailedEvent += value;
            remove => _startFailedEvent -= value;
        }
        private EventHandler<StartFailedEventArgs> _startFailedEvent;
        private void OnStartFailed(StartFailedEventArgs e)
        {
            _startFailedEvent?.Invoke(this, e);
        }

        /// <summary>
        /// Tritt auf, wenn eine Verbindung zum Service hergestellt wurde.
        /// </summary>
        internal event EventHandler<ConnectedEventArgs> Connected
        {
            add => _connectedEvent += value;
            remove => _connectedEvent -= value;
        }
        private EventHandler<ConnectedEventArgs> _connectedEvent;
        private void OnConnected(ConnectedEventArgs e)
        {
            _connectedEvent?.Invoke(this, e);
        }
        /// <summary>
        /// Tritt auf, wenn ein Scanner mit diesem Empfänger verbunden wurde.
        /// </summary>
        internal event EventHandler<ScannerConnectedEventArgs> ScannerConnected
        {
            add => _scannerConnectedEvent += value;
            remove => _scannerConnectedEvent -= value;
        }
        private EventHandler<ScannerConnectedEventArgs> _scannerConnectedEvent;
        private void OnScannerConnected(ScannerConnectedEventArgs e)
        {
            _scannerConnectedEvent?.Invoke(this, e);
        }
        /// <summary>
        /// Tritt auf, wenn ein mit diesem Empfänger verbundener Scanner seine Verbindung getrennt hat.
        /// <remarks>Kann beispielsweise duch Trennung der Internetverbindung oder duch Schliessen der App auftreten.</remarks>
        /// </summary>
        internal event EventHandler<EventArgs> ScannerDisconnected
        {
            add => _scannerDisconnectedEvent += value;
            remove => _scannerDisconnectedEvent -= value;
        }
        private EventHandler<EventArgs> _scannerDisconnectedEvent;
        private void OnScannerDisconnected(EventArgs e)
        {
            _scannerDisconnectedEvent?.Invoke(this, e);
        }
        /// <summary>
        /// Tritt auf, wenn der Scanner Daten gelesen hat.
        /// </summary>
        internal event EventHandler<ScannedEventArgs> Scanned
        {
            add => _scannedEvent += value;
            remove => _scannedEvent -= value;
        }
        private EventHandler<ScannedEventArgs> _scannedEvent;
        private void OnScanned(ScannedEventArgs e)
        {
            _scannedEvent?.Invoke(this, e);
        }
        /// <summary>
        /// Tritt auf, wenn eine Verbindung zum Service geschlossen wurde.
        /// </summary>
        internal event EventHandler<EventArgs> Close
        {
            add => _closeEvent += value;
            remove => _closeEvent -= value;
        }
        private EventHandler<EventArgs> _closeEvent;
        private void OnClose(EventArgs e)
        {
            _closeEvent?.Invoke(this, e);
        }
        /// <summary>
        /// Tritt auf, wenn beim Service einen Fehler aufgetreten ist
        /// </summary>
        internal event EventHandler<ServerErrorEventArgs> ServerError
        {
            add => _serverErrorEvent += value;
            remove => _serverErrorEvent -= value;
        }
        private EventHandler<ServerErrorEventArgs> _serverErrorEvent;
        private void OnServerError(ServerErrorEventArgs e)
        {
            _serverErrorEvent?.Invoke(this, e);
        }
        #endregion Events
    }
}
